import argparse
import os
import random
import warnings
from collections import defaultdict
from typing import Counter, Sequence

import fsspec
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import torch
import torch.distributions as td
from tensordict import TensorDict

from causica.data_generation.samplers.functional_relationships_sampler import (
    FunctionalRelationshipsSampler,
    LinearRelationshipsSampler,
    RFFFunctionalRelationshipsSampler,
)
from causica.data_generation.samplers.noise_dist_sampler import (
    BernoulliNoiseModuleSampler,
    CategoricalNoiseModuleSampler,
    JointNoiseModuleSampler,
    NoiseModuleSampler,
    UnivariateNormalNoiseModuleSampler,
)
from causica.data_generation.samplers.sem_sampler import SEMSampler
from causica.datasets.causal_dataset import CausalDataset
from causica.datasets.causica_dataset_format import (
    CounterfactualWithEffects,
    InterventionWithEffects,
    Variable,
    VariablesMetadata,
    get_group_variable_names,
    save_dataset,
)
from causica.datasets.interventional_data import CounterfactualData, InterventionData
from causica.datasets.tensordict_utils import expand_tensordict_groups
from causica.datasets.variable_types import VariableTypeEnum
from causica.distributions.adjacency import AdjacencyDistribution, ErdosRenyiDAGDistribution, FixedAdjacencyDistribution
from causica.sem.structural_equation_model import SEM

try:
    import seaborn as sns
except ImportError:
    sns = None


def get_graph_distribution(
    graph_type: str,
    num_nodes: int | None = None,
    num_edges: int | None = None,
    probs: float | None = None,
    graph_file: str | None = None,
    **storage_options,
) -> AdjacencyDistribution:
    """Get a graph from a graph type.

    Args:
        graph_type: The type of graph to generate. Either "er" for Erdos-Renyi or "numpy" for a numpy graph.
        num_nodes: The number of nodes in the graph. Not used if using a numpy graph.
        num_edges: The number of edges in the graph. Not used if using a numpy graph.
        probs: A float of the probability that an edge exists between 2 nodes
        graph_file: The path to a graph file if using a numpy graph.
        storage_options: The storage options to pass to fsspec.

    Returns:
        The adjacency matrix of the graph.
    """
    if graph_type == "er":
        if num_nodes is None:
            raise ValueError("Must provide num_nodes for Erdos-Renyi graph.")
        num_edges_tensor = None
        if num_edges is not None:
            num_edges_tensor = torch.tensor(num_edges, dtype=torch.float32)
        probs_tensor = None
        if probs is not None:
            probs_tensor = torch.tensor(probs)
        return ErdosRenyiDAGDistribution(num_nodes=num_nodes, num_edges=num_edges_tensor, probs=probs_tensor)
    if graph_type == "numpy":
        if graph_file is None:
            raise ValueError("Must provide graph_file for numpy graph.")
        with fsspec.open(graph_file, **storage_options) as f:
            return FixedAdjacencyDistribution(torch.tensor(np.load(f)))

    raise ValueError("Unknown graph type.")


def get_functional_relationship_sampler(
    function_type: str, shapes_dict: dict[str, torch.Size]
) -> FunctionalRelationshipsSampler:
    """Get a functional relationship from a function type.

    Args:
        function_type: The type of function to generate. Either "linear" or "rff".
        shapes_dict: A dictionary mapping variable names to their shapes.

    Returns:
        The functional relationship.
    """
    dim = sum(s.numel() for s in shapes_dict.values())
    if function_type == "linear":
        ones_matrix = torch.ones((dim, dim), dtype=torch.float32)
        functional_relationships_sampler: FunctionalRelationshipsSampler = LinearRelationshipsSampler(
            scale_dist=td.Uniform(
                low=ones_matrix,
                high=3.0 * ones_matrix,
            ),
            shapes_dict=shapes_dict,
        )
    elif function_type == "rff":
        num_rf = 100
        ones_matrix = torch.ones((num_rf, dim), dtype=torch.float32)
        ones = torch.ones((num_rf,), dtype=torch.float32)
        functional_relationships_sampler = RFFFunctionalRelationshipsSampler(
            rf_dist=td.Uniform(
                low=7.0 * ones_matrix,
                high=10.0 * ones_matrix,
            ),
            coeff_dist=td.Uniform(
                low=10.0 * ones,
                high=20.0 * ones,
            ),
            shapes_dict=shapes_dict,
        )
    else:
        raise ValueError(f"Unknown function type. Got {function_type}")

    return functional_relationships_sampler


def get_noise_module_sampler(variables: VariablesMetadata) -> JointNoiseModuleSampler:
    """Get a noise module sampler from a list of variables.

    Args:
        variables: Variable specifications to generate noise modules for.

    Returns:
        The noise module sampler.
    """
    size_dict = get_size_dict(variables)
    variable_type_dict = get_variable_type_dict(variables)

    noise_dist_samplers: dict[str, NoiseModuleSampler] = {}
    for group, size in size_dict.items():
        if variable_type_dict[group] == VariableTypeEnum.BINARY:
            noise_dist_samplers[group] = BernoulliNoiseModuleSampler(
                base_logits_dist=td.Uniform(low=-5, high=5), dim=size
            )
        elif variable_type_dict[group] == VariableTypeEnum.CATEGORICAL:
            noise_dist_samplers[group] = CategoricalNoiseModuleSampler(
                base_logits_dist=td.Uniform(low=-5, high=5),
                num_classes=size,
            )
        else:
            noise_dist_samplers[group] = UnivariateNormalNoiseModuleSampler(
                std_dist=td.Uniform(low=0.2, high=2.0), dim=size
            )

    return JointNoiseModuleSampler(noise_dist_samplers)


def get_variable_type_dict(
    variables: VariablesMetadata,
) -> dict[str, VariableTypeEnum]:
    """Get a dictionary mapping variable groups to their types.

    This also ensures that every variable in a group has the same type.

    Args:
        variables: Variable specifications to get the types for.

    Returns:
        The variable type dictionary.
    """
    variable_type_dict = defaultdict(list)
    for variable in variables.variables:
        variable_type_dict[variable.group_name].append(variable.type)

    for group, types in variable_type_dict.items():
        assert len(set(types)) == 1, f"Found multiple variable types in group {group}"

        if types[0] == VariableTypeEnum.CATEGORICAL:
            assert len(types) == 1, "Groups of categorical variables must have only one variable."

    return {k: v[0] for k, v in variable_type_dict.items()}


def get_size_dict(variables: VariablesMetadata) -> dict[str, int]:
    """Get a dictionary mapping variable groups to their sizes.

    Args:
        variables: Variable specifications to get the sizes for.

    Returns:
        The size dictionary.
    """
    size_dict = Counter(d.group_name for d in variables.variables)
    for variable in variables.variables:
        if variable.type == VariableTypeEnum.CATEGORICAL:
            assert variable.lower is not None and variable.upper is not None
            difference = variable.upper - variable.lower + 1
            assert difference // 1 == difference, ""
            assert size_dict[variable.group_name] == 1
            size_dict[variable.group_name] = int(difference)

    return size_dict


def sample_treatment_given_effects(graph: torch.Tensor, node_names: Sequence[str], effect_variables: list[str]) -> str:
    """Sample a treatment from the graph.

    Args:
        graph: The adjacency matrix of the graph.
        node_names: The names of the nodes in the graph.
        effect_variables: pre-specified effect variables.

    Returns:
        The treatment variable.
    """
    nx_graph = nx.from_numpy_array(graph.numpy(), create_using=nx.DiGraph)
    nx_graph = nx.relabel_nodes(nx_graph, dict(enumerate(node_names)))

    common_ancestors = nx.ancestors(nx_graph, effect_variables[0])
    for effect_variable in effect_variables[1:]:
        common_ancestors &= nx.ancestors(nx_graph, effect_variable)
    if common_ancestors:
        treatment = np.random.choice(list(common_ancestors), size=1, replace=False).item()
        return treatment
    warnings.warn("No common ancestors found for the given effects. Returning random sampled treatment.")
    possible_treatments = [node for node in node_names if node not in effect_variables]
    treatment = np.random.choice(possible_treatments, size=1, replace=False).item()
    return treatment


def sample_effects_given_treatment(
    graph: torch.Tensor,
    node_names: Sequence[str],
    treatment_variable: str,
    num_effects: int = 1,
) -> list[str]:
    """Sample effect variables from the graph given a treatment.

    Args:
        graph: The adjacency matrix of the graph.
        node_names: The names of the nodes in the graph.
        treatment_variable: The treatment variable from which to find effects.
        num_effects: The number of effect nodes to sample.

    Returns:
        The list of effect variables.
    """
    nx_graph = nx.from_numpy_array(graph.numpy(), create_using=nx.DiGraph)
    nx_graph = nx.relabel_nodes(nx_graph, dict(enumerate(node_names)))

    descendants = list(nx.descendants(nx_graph, treatment_variable))
    if not descendants:
        warnings.warn("No descendants found for the given treatment. Defaulting to random sampling.")
        possible_effects = [node for node in node_names if node != treatment_variable]
        return np.random.choice(possible_effects, size=num_effects, replace=False).tolist()

    return np.random.choice(descendants, size=num_effects, replace=False).tolist()


def sample_treatment_and_effect(
    graph: torch.Tensor,
    node_names: Sequence[str],
    ensure_effect: bool = True,
    num_effects: int = 1,
) -> tuple[str, list[str]]:
    """Sample a treatment and effects from a graph.

    Args:
        graph: The adjacency matrix of the graph.
        node_names: The names of the nodes in the graph.
        ensure_effect: Whether to ensure that there is a path from the treatment to the effect.
        num_effects: The number of effect nodes to sample.
    Returns:
        A tuple containing the treatment variable and a list of effect variables.
    """

    if ensure_effect:
        nx_graph = nx.from_numpy_array(graph.numpy(), create_using=nx.DiGraph)
        nx_graph = nx.relabel_nodes(nx_graph, dict(enumerate(node_names)))
        nodes_with_outgoing_edges = [n for n in node_names if any(True for _ in nx_graph.successors(n))]
        if nodes_with_outgoing_edges:
            treatment = np.random.choice(nodes_with_outgoing_edges, size=1, replace=False).item()
            descendants = list(nx.descendants(nx_graph, treatment))
            effects = np.random.choice(descendants, size=num_effects, replace=False).tolist()
            return treatment, effects
        warnings.warn("No nodes with descendants found in the graph. Defaulting to random sampling.")

    samples = np.random.choice(node_names, size=1 + num_effects, replace=False)
    treatment = samples[0]
    effects = samples[1:].tolist()

    return treatment, effects


def sample_dataset(
    sem: SEM,
    sample_dataset_size: torch.Size,
    num_interventions: int = 0,
    num_intervention_samples: int = 1000,
    sample_interventions: bool = False,
    sample_counterfactuals: bool = False,
    treatment_variable: str | None = None,
    effect_variables: list[str] | None = None,
) -> CausalDataset:
    """Sample a new dataset and returns it as a CausalDataset object.

    Args:
        sem: The SEM to sample from.
        sample_dataset_size: The size of the dataset to sample from the SEM
        num_interventions: The number of interventions to sample per dataset. If 0, no interventions are sampled.
        num_intervention_samples: The number of interventional samples to sample.
        sample_interventions: Whether to sample interventions.
        sample_counterfactuals: Whether to sample counterfactuals.
        treatment_variable: pre-specified treatment
        effet_variables: pre-specified effects
    Returns:
        A CausalDataset object holding the data, graph and potential interventions and counterfactuals.
    """

    noise = sem.sample_noise(sample_dataset_size)
    observations = sem.noise_to_sample(noise)

    interventions: list[InterventionWithEffects] = []
    counterfactuals: list[CounterfactualWithEffects] = []

    for _ in range(num_interventions):
        if treatment_variable and effect_variables:
            treatment, effects = treatment_variable, effect_variables
        elif treatment_variable:
            treatment, effects = treatment_variable, sample_effects_given_treatment(
                sem.graph, sem.node_names, treatment_variable
            )
        elif effect_variables:
            treatment, effects = (
                sample_treatment_given_effects(sem.graph, sem.node_names, effect_variables),
                effect_variables,
            )
        else:
            treatment, effects = sample_treatment_and_effect(sem.graph, sem.node_names)

        if sample_interventions:
            interventions.append(
                (
                    sample_intervention(sem, observations, num_intervention_samples, treatment=treatment),
                    sample_intervention(sem, observations, num_intervention_samples, treatment=treatment),
                    set(effects),
                )
            )

        if sample_counterfactuals:
            cf_noise = sem.sample_noise(
                torch.Size(
                    [
                        num_intervention_samples,
                    ]
                )
            )
            cf_observations = sem.noise_to_sample(cf_noise)

            counterfactuals.append(
                (
                    sample_counterfactual(sem, cf_observations, cf_noise, treatment=treatment),
                    sample_counterfactual(sem, cf_observations, cf_noise, treatment=treatment),
                    set(effects),
                )
            )

    return CausalDataset(
        observations=observations,
        noise=noise,
        graph=sem.graph,
        interventions=interventions,
        counterfactuals=counterfactuals,
    )


def sample_intervention_dict(tensordict_data: TensorDict, treatment: str | None = None) -> TensorDict:
    """Sample an intervention from a given SEM.

    This samples a random value for the treatment variable from the data. The value is sampled uniformly from the
    range of the treatment variable in the data.

    The treatment variable is chosen randomly across all nodes if not specified.

    Args:
        tensordict_data: Base data for sampling an intervention value.
        treatment: The name of the treatment variable. If None, a random variable is chosen across the tensordict keys.

    Returns:
        A TensorDict holding the intervention value.
    """
    if treatment is None:
        treatment = random.choice(list(tensordict_data.keys()))

    batch_axes = tuple(range(tensordict_data.batch_dims))
    treatment_shape = tensordict_data[treatment].shape[tensordict_data.batch_dims :]
    treatment_max = torch.amax(tensordict_data[treatment], dim=batch_axes)
    treatment_min = torch.amin(tensordict_data[treatment], dim=batch_axes)

    treatment_a = torch.rand(treatment_shape) * (treatment_max - treatment_min) + treatment_min

    return TensorDict({treatment: treatment_a}, batch_size=torch.Size())


def sample_intervention(
    sem: SEM, tensordict_data: TensorDict, num_intervention_samples: int, treatment: str | None = None
) -> InterventionData:
    """Sample an intervention and it's sample mean from a given SEM.

    Args:
        sem: SEM to sample interventional data from.
        tensordict_data: Base data for sampling an intervention value.
        num_intervention_samples: The number of samples to draw from the interventional distribution.
        treatment: The name of the treatment variable. If None, a random variable is chosen.

    Returns:
        an intervention data object
    """
    intervention_a = sample_intervention_dict(tensordict_data, treatment=treatment)

    intervention_a_samples = sem.do(intervention_a).sample(
        torch.Size(
            [
                num_intervention_samples,
            ]
        )
    )

    return InterventionData(
        intervention_a_samples,
        intervention_a,
        TensorDict({}, batch_size=torch.Size()),
    )


def sample_counterfactual(
    sem: SEM, factual_data: TensorDict, noise: TensorDict, treatment: str | None = None
) -> CounterfactualData:
    """Sample an intervention and it's sample mean from a given SEM.

    Args:
        sem: SEM to sample counterfactual data from.
        factual_data: Base data for sampling an counterfactual value.
        noise: Base noise for sampling an counterfactual value.
        treatment: The name of the treatment variable. If None, a random variable is chosen.

    Returns:
        an counterfactual data object
    """
    intervention_a = sample_intervention_dict(factual_data, treatment=treatment)

    counterfactuals = sem.do(intervention_a).noise_to_sample(noise)

    return CounterfactualData(counterfactuals, intervention_a, factual_data)


def get_variable_definitions(
    variable_json_path: str | None = None,
    num_nodes: int | None = None,
    storage_options: dict[str, str] | None = None,
) -> VariablesMetadata:
    """Get the variable definitions.

    Args:
        variable_json_path: The path to a json file containing the variables. This can be any fsspec compatible url.
        num_nodes: The number of nodes in the graph. Not used if using a numpy graph.
        storage_options: The storage options to pass to fsspec.

    Returns:
        The variables.
    """
    if storage_options is None:
        storage_options = {}

    if variable_json_path:
        with fsspec.open(variable_json_path, **storage_options) as f:
            variables = VariablesMetadata.from_json(f.read())  # type: ignore
    elif num_nodes is not None:
        variables = VariablesMetadata([Variable(f"x_{i}", f"x_{i}") for i in range(num_nodes)])
    else:
        raise ValueError("Must provide either variable_json_path or num_nodes.")

    return variables


def generate_sem_sampler(
    variables: VariablesMetadata,
    graph_file: str | None,
    num_edges: int | None,
    graph_type: str,
    function_type: str,
    probs: float | None = None,
    storage_options: dict[str, str] | None = None,
) -> SEMSampler:
    """Generates a SEM according to specifications

    Args:
        variable_json_path: The path to a json file containing the variables. This can be any fsspec compatible url.
        num_nodes: The number of nodes in the graph. Not used if using a numpy graph.
        graph_file: The path to a graph file if using a numpy graph. This can be any fsspec compatible url.
        num_edges: The number of edges in the graph. Not used if using a numpy graph.
        probs: A float of the probability that an edge exists between 2 nodes
        graph_type: The type of graph to generate. Either "er" for Erdos-Renyi or "numpy" for a numpy graph.
        function_type: The type of function to generate. Either "linear" or "rff".
        storage_options: The storage options to pass to fsspec.

    Returns:
        The variables and SEM.
    """
    if storage_options is None:
        storage_options = {}

    size_dict = get_size_dict(variables)
    shapes_dict = {k: torch.Size((s,)) for k, s in size_dict.items()}
    num_nodes = len(size_dict)

    print("Generating SEM sampler...")
    adjacency_dist = get_graph_distribution(graph_type, num_nodes, num_edges, probs, graph_file, **storage_options)
    functional_relationships_sampler = get_functional_relationship_sampler(function_type, shapes_dict)
    joint_noise_dist_sampler = get_noise_module_sampler(variables)

    sem_sampler = SEMSampler(
        adjacency_dist=adjacency_dist,
        joint_noise_module_sampler=joint_noise_dist_sampler,
        functional_relationships_sampler=functional_relationships_sampler,
    )

    return sem_sampler


def plot_dataset(
    data: TensorDict,
    variables: VariablesMetadata,
    plot_kind: str = "kde",
    plot_num: int = 10,
    datadir: str = "",
    storage_options: dict[str, str] | None = None,
) -> None:
    """Plot the data distribution.

    Args:
        data: The data to plot.
        variables: Variable specifications.
        plot_kind: Type of joint plot to create.
        plot_num: Maximum number of variables to plot.
        datadir: The directory to save the dataset to. This can be any fsspec compatible url.
        storage_options: The storage options to pass to fsspec.
    """
    if sns is None:
        raise ImportError("Please install seaborn to plot the data distribution.")

    if storage_options is None:
        storage_options = {}

    group_names = get_group_variable_names(variables)
    variable_types = get_variable_type_dict(variables)
    data.update_(
        {
            k: torch.argmax(v, dim=-1, keepdim=False)
            for k, v in data.items()
            if variable_types[k] == VariableTypeEnum.CATEGORICAL
        }
    )
    data_dict = {k: v.numpy().squeeze() for k, v in expand_tensordict_groups(data, group_names).to_dict().items()}
    if len(data_dict) > plot_num:
        print(f"Too many variables ({len(data_dict)}) to plot. Only plotting the first {plot_num}...")

    df = pd.DataFrame(data=data_dict)
    df = df.iloc[:, :plot_num]
    sns.pairplot(data=df, kind=plot_kind)
    img_path = os.path.join(datadir, "data_distribution.png")
    with fsspec.open(img_path, mode="wb", **storage_options) as f:
        plt.savefig(f)
        print(f"Saved data distribution plot to {img_path}")


def generate_save_plot_synthetic_data(
    graph_type: str,
    function_type: str,
    datadir: str,
    num_samples_train: int,
    num_samples_test: int,
    num_interventions: int,
    variable_json_path: str | None = None,
    num_nodes: int | None = None,
    graph_file: str | None = None,
    num_edges: int | None = None,
    probs: float | None = None,
    overwrite: bool = False,
    plot_kind: str = "",
    plot_num: int = 10,
    storage_options: dict[str, str] | None = None,
) -> None:
    """Generate, save and plot synthetic data.

    This will sample a SEM from the given specifications, generate data from it, and save it to disk following the
    Causica dataset format. It optionally plots the joint distribution over the variables.


    Args:
        graph_type: The type of graph to generate. Either "er" for Erdos-Renyi or "numpy" for a numpy graph.
        function_type: The type of function to generate. Either "linear" or "rff".
        datadir: The directory to save the dataset to. This can be any fsspec compatible url.
        num_samples_train: The number of training samples to generate.
        num_samples_test: The number of validation and test samples to generate.
        num_interventions: The number of interventions and counterfactuals to generate.
        variable_json_path: The path to a json file containing the variables. This can be any fsspec compatible url.
        num_nodes: The number of nodes in the graph. Not used if using a numpy graph.
        graph_file: The path to a graph file if using a numpy graph. This can be any fsspec compatible url.
        num_edges: The number of edges in the graph. Not used if using a numpy graph.
        probs: A float of the probability that an edge exists between 2 nodes
        overwrite: Whether to overwrite the dataset if it already exists.
        plot_kind: Type of joint plot to create.
        plot_num: Maximum number of variables to plot.
        storage_options: The storage options to pass to fsspec.
    """
    if storage_options is None:
        storage_options = {}

    variables = get_variable_definitions(variable_json_path, num_nodes, storage_options)
    sem_sampler = generate_sem_sampler(
        variables=variables,
        graph_file=graph_file,
        num_edges=num_edges,
        probs=probs,
        graph_type=graph_type,
        function_type=function_type,
        storage_options=storage_options,
    )
    sem = sem_sampler.sample()

    dataset = sample_dataset(
        sem=sem,
        sample_dataset_size=torch.Size([num_samples_train + num_samples_test * 2]),
        num_interventions=num_interventions,
        num_intervention_samples=num_samples_test,
        sample_interventions=True,
        sample_counterfactuals=True,
    )

    print(f"Saving dataset to {datadir}...")
    save_dataset(
        datadir,
        variables,
        sem.graph,
        dataset.observations[:num_samples_train],
        dataset.observations[num_samples_train : num_samples_train + num_samples_test],
        dataset.observations[num_samples_train + num_samples_test :],
        dataset.interventions,
        dataset.counterfactuals,
        overwrite=overwrite,
        **storage_options,
    )

    if plot_kind and plot_num > 0:
        plot_dataset(dataset.observations, variables, plot_kind, plot_num, datadir, storage_options)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--variable-json", "-v", type=str)
    parser.add_argument("--num-samples-train", "-n", type=int, default=1000)
    parser.add_argument("--num-samples-test", "-t", type=int, default=1000)
    parser.add_argument("--num-interventions", "-i", type=int, default=1)
    parser.add_argument("--num-variables", "-nv", type=int)
    parser.add_argument("--num-edges", "-ne", type=int)
    parser.add_argument("--edge-probabilities", "-p", type=float)
    parser.add_argument(
        "--graph-type",
        "-g",
        type=str.lower,
        default="er",
        choices=["er", "numpy"],
        help="One of er or numpy to load fixed adj matrix in npy format.",
    )
    parser.add_argument("--graph_file", "-gf", type=str)
    parser.add_argument("--seed", "-s", type=int)
    parser.add_argument("--datadir", "-d", type=str)
    parser.add_argument(
        "--function-type",
        "-f",
        type=str.lower,
        default="linear",
        choices=["linear", "rff"],
    )
    parser.add_argument("--overwrite", action="store_true")
    parser.add_argument("--plot-kind", type=str, default="kde")
    parser.add_argument("--plot-num", type=int, default=10)
    parser.add_argument(
        "--storage-options",
        type=str,
        action="append",
        help="Storage options for fsspec. Options should be in the form key=value. For example --storage-options 'anon=True'",
    )

    args = parser.parse_args()

    if args.seed is not None:
        np.random.seed(args.seed)
        random.seed(args.seed)
        torch.manual_seed(args.seed)

    storage_options = None
    if args.storage_options:
        storage_options = {
            k: v.replace('"', "").replace("'", "").strip() for t in args.storage_options for k, v in [t.split("=")]
        }

        for k, v in storage_options.items():
            if v.lower() in ("true", "false"):
                storage_options[k] = v.lower() == "true"

    generate_save_plot_synthetic_data(
        variable_json_path=args.variable_json,
        num_nodes=args.num_variables,
        graph_file=args.graph_file,
        num_edges=args.num_edges,
        probs=args.edge_probabilities,
        graph_type=args.graph_type,
        datadir=args.datadir,
        function_type=args.function_type,
        num_samples_train=args.num_samples_train,
        num_samples_test=args.num_samples_test,
        num_interventions=args.num_interventions,
        overwrite=args.overwrite,
        plot_kind=args.plot_kind,
        plot_num=args.plot_num,
        storage_options=storage_options,
    )


if __name__ == "__main__":
    main()



import math
from typing import Optional

import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import FunctionalRelationships


class RFFFunctionalRelationships(FunctionalRelationships):
    """
    A simple random fourier feature-based functional relationship.

    The formula implemented here is:
    x_i = sqrt(2/M) * output_scales_i sum_{i}^{M} alpha_i sin( <w_i, pa(x_i)> / length_scale_i )
    """

    def __init__(
        self,
        shapes: dict[str, torch.Size],
        initial_random_features: torch.Tensor,
        initial_coefficients: torch.Tensor,
        initial_bias: Optional[torch.Tensor] = None,
        initial_length_scales: Optional[torch.Tensor] = None,
        initial_output_scales: Optional[torch.Tensor] = None,
        initial_angles: Optional[torch.Tensor] = None,
        trainable: bool = False,
    ) -> None:
        """
        Args:
            shapes: Dict of node shapes (how many dimensions a variable has)
                Order corresponds to the order in graph(s).
            initial_random_features: a tensor containing the random features [num_rf, output_shape]
            initial_coefficients: a tensor containing the linear outer coefficients [num_rf,]
            initial_bias: Optional, None or a tensor containing the bias [output_shape,]
            initial_length_scales: Optional, None or a tensor containing the length scales [output_shape,]
            initial_output_scales: Optional, None or a tensor containing the output scales [output_shape,]
            initial_angles: Optional, None or a tensor containing the angles [num_rf,]
            trainable: whether the coefficient matrix should be learnable
        """
        super().__init__(shapes=shapes)

        assert initial_random_features.shape[0] == initial_coefficients.shape[0]
        self.num_rf = initial_random_features.shape[0]

        self.shape = self.tensor_to_td.output_shape
        assert initial_random_features.shape[1] == self.shape

        if initial_length_scales is not None:
            assert initial_length_scales.shape[0] == self.shape
        else:
            initial_length_scales = torch.ones(self.shape)

        if initial_output_scales is not None:
            assert initial_output_scales.shape[0] == self.shape
        else:
            initial_output_scales = torch.ones(self.shape)

        if initial_bias is not None:
            assert initial_bias.shape[0] == self.shape
        else:
            initial_bias = torch.zeros(self.shape)

        if initial_angles is not None:
            assert initial_angles.shape[0] == self.num_rf
        else:
            initial_angles = (-math.pi / 2) * torch.ones(self.num_rf)

        self.bias = torch.nn.Parameter(initial_bias, requires_grad=trainable)
        self.length_scales = torch.nn.Parameter(initial_length_scales, requires_grad=trainable)
        self.output_scales = torch.nn.Parameter(initial_output_scales, requires_grad=trainable)
        self.angles = torch.nn.Parameter(initial_angles, requires_grad=trainable)
        self.linear_coefficients_inner = torch.nn.Parameter(initial_random_features, requires_grad=trainable)
        self.linear_coefficients_outer = torch.nn.Parameter(initial_coefficients, requires_grad=trainable)

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """
        Args:
            samples: Batched inputs, size batch_size_x + batch_shape_g + (concatenated_shape).
            graphs: Weighted adjacency matrix, size batch_size_g + (n, n)
        Returns:
            A Dict of tensors of shape batch_shape_x + batch_shape_g + (concatenated_shape)
        """
        return self.tensor_to_td(self.non_linear_map(self.tensor_to_td.inv(samples), graphs))

    def non_linear_map(self, samples: torch.Tensor, graph: torch.Tensor) -> torch.Tensor:
        """
        Applies the non linear function to a concatenated tensor of samples.

        Args:
            samples: tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
            graph: tensor of shape batch_shape_g + [n_nodes, n_nodes]
        Returns:
            tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
        """
        batch_shape_samples = samples.shape[:-1]
        batch_shape_g = graph.shape[:-2]
        batch_shape_x_f = batch_shape_samples[: -len(batch_shape_g)]
        if len(batch_shape_g) > 0 and batch_shape_samples[-len(batch_shape_g) :] != batch_shape_g:
            raise ValueError(
                f"Batch shape of samples and graph must match but got {batch_shape_samples} and {batch_shape_g}"
            )

        masked_graph = torch.einsum("ji,...jk,kl->...il", self.stacked_key_masks, graph, self.stacked_key_masks)

        graph_broad = masked_graph.expand(*(batch_shape_x_f + (-1,) * len(masked_graph.shape)))

        inner_prods = torch.einsum("ij,...j,...jk->...ik", self.linear_coefficients_inner, samples, graph_broad)
        rescaled_inner_prods = inner_prods / self.length_scales
        rescaled_and_translated_inner_prods = rescaled_inner_prods + self.angles[..., None]
        return self.bias + (
            math.sqrt(2 / self.num_rf)
            * self.output_scales
            * torch.einsum(
                "i,...ij->...j", self.linear_coefficients_outer, torch.cos(rescaled_and_translated_inner_prods)
            )
        )



from typing import Optional

import torch

from causica.functional_relationships.rff_functional_relationships import RFFFunctionalRelationships


class HeteroscedasticRFFFunctionalRelationships(RFFFunctionalRelationships):
    """
    A simple random fourier feature-based functional relationship.

    The formula implemented here is:
    x_i = sqrt(2/M) * output_scales_i sum_{i}^{M} alpha_i sin( <w_i, pa(x_i)> / length_scale_i)
    """

    def __init__(
        self,
        shapes: dict[str, torch.Size],
        initial_random_features: torch.Tensor,
        initial_coefficients: torch.Tensor,
        initial_bias: Optional[torch.Tensor] = None,
        initial_length_scales: Optional[torch.Tensor] = None,
        initial_output_scales: Optional[torch.Tensor] = None,
        initial_angles: Optional[torch.Tensor] = None,
        trainable: bool = False,
        log_scale: bool = False,
    ) -> None:
        """
        Args:
            shapes: Dict of node shapes (how many dimensions a variable has)
                Order corresponds to the order in graph(s).
            initial_random_features: a tensor containing the random features [num_rf, output_shape]
            initial_coefficients: a tensor containing the linear outer coefficients [num_rf,]
            initial_bias: Optional, None or a tensor containing the bias [output_shape,]
            initial_length_scales: Optional, None or a tensor containing the length scales [output_shape,]
            initial_output_scales: Optional, None or a tensor containing the output scales [output_shape,]
            initial_angles: Optional, None or a tensor containing the angles [num_rf,]
            trainable: whether the coefficient matrix should be learnable
            log_scale: whether to apply a log scale to the output
        """
        super().__init__(
            shapes=shapes,
            initial_random_features=initial_random_features,
            initial_coefficients=initial_coefficients,
            initial_bias=initial_bias,
            initial_length_scales=initial_length_scales,
            initial_output_scales=initial_output_scales,
            initial_angles=initial_angles,
            trainable=trainable,
        )
        self.log_scale = log_scale

    def non_linear_map(self, samples: torch.Tensor, graph: torch.Tensor) -> torch.Tensor:
        """
        Applies the non linear function to a concatenated tensor of samples.

        Args:
            samples: tensor of shape batch_shape_x + [n_cols]
            graph: tensor of shape batch_shape_g + [n_nodes, n_nodes]
        Returns:
            tensor of shape batch_shape_x + batch_shape_g + [n_cols]
        """
        res = super().non_linear_map(samples, graph)
        res = torch.log(1.0 + torch.exp(res))
        if self.log_scale:
            res = torch.log(torch.log(1.0 + torch.exp(res)))
        return res



from typing import Optional

import torch
from torch import nn

from causica.nn.fip.attention_layer import MultiHeadAttention
from causica.nn.fip.embeddings import PositionWiseFeedForward


class CausalDecoderLayer(nn.Module):
    """
    Causal Decoder Layer.

    It consists of a multi-head self-attention mechanism, followed by a feed-forward network.
    It also allows the model to use the causal attention mechanism.

    """

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_key: int,
        d_ff: int,
        dropout: float,
        max_seq_length: int,
        attn_type: str = "standard",
        cost_type: str = "dot_product",
    ):
        super().__init__()

        self.self_attn = MultiHeadAttention(
            d_model, num_heads, max_seq_length, dim_key=dim_key, attn_type=attn_type, cost_type=cost_type
        )

        self.feed_forward = PositionWiseFeedForward(d_model, d_ff)
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.max_seq_length = max_seq_length

    def forward(
        self, queries: torch.Tensor, keys: torch.Tensor, values: torch.Tensor, mask: Optional[torch.Tensor]
    ) -> torch.Tensor:

        """
        Args:
            queries: expected shape: (batch_size, max_seq_length, d_model)
            keys: expected shape: (batch_size, max_seq_length, d_model)
            values: expected shape: (batch_size, max_seq_length, d_model)
            mask: expected shape: (batch_size, max_seq_length, max_seq_length)

        Returns:
            output: expected shape: (batch_size, max_seq_length, d_model)

        """

        attn_output = self.self_attn(queries, keys, values, mask)
        x_trans = self.norm1(queries + self.dropout(attn_output))

        ff_output = self.feed_forward(x_trans)

        return self.norm2(x_trans + self.dropout(ff_output))


class AmortizedDecoderLayer(nn.Module):
    """
    Amortized Causal Decoder Layer.

    This is a re-implementation of the decoder layer used in the AVICI model in PyTorch:

    https://github.com/larslorch/avici

    """

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_key: int,
        d_ff: int,
        dropout: float,
        max_seq_length: int,
        attn_type: str = "standard",
        cost_type: str = "dot_product",
    ):
        super().__init__()

        self.norm_q = nn.LayerNorm(d_model)
        self.norm_k = nn.LayerNorm(d_model)
        self.norm_v = nn.LayerNorm(d_model)

        self.self_attn = MultiHeadAttention(
            d_model, num_heads, max_seq_length, dim_key=dim_key, attn_type=attn_type, cost_type=cost_type
        )

        self.feed_forward = PositionWiseFeedForward(d_model, d_ff)

        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

    def forward(
        self, queries: torch.Tensor, keys: torch.Tensor, values: torch.Tensor, mask: Optional[torch.Tensor]
    ) -> torch.Tensor:

        """
        Args:
            queries: expected shape: (batch_size, max_seq_length, d_model)
            keys: expected shape: (batch_size, max_seq_length, d_model)
            values: expected shape: (batch_size, max_seq_length, d_model)

        Returns:
            output: expected shape: (batch_size, max_seq_length, d_model)

        """
        queries_in = self.norm_q(queries)
        keys_in = self.norm_k(keys)
        values_in = self.norm_v(values)
        attn_output = self.self_attn(queries_in, keys_in, values_in, mask)
        x = queries + self.dropout(attn_output)
        x_in = self.norm(x)
        ff_output = self.dropout(self.feed_forward(x_in))
        return x + ff_output



import igraph as ig
import numpy as np
import torch

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution


class StochasticBlockModelDAGDistribution(AdjacencyDistribution):
    """
    An adjacency distribution for sampling Directed Acyclic Graphs using SBM.

    Re-implementation in pytorch from AVICI:

    https://github.com/larslorch/avici


    """

    def __init__(
        self,
        num_nodes: int,
        edges_per_node: list[int],
        num_blocks: list[int],
        damping: list[float],
        validate_args: bool | None = None,
    ):
        """
        Args:
            num_nodes: the number of nodes in the DAGs to be sampled
            edges_per_node: List of the number of edges per node to sample.
            num_blocks: List of the number of blocks in the model
            damp: List of the damp factor for inter block edges. damp = 1.0 is equivalent to erdos renyi
            validate_args: Optional arguments from AdjacencyDistribution
        """
        assert num_nodes > 0, "Number of nodes in the graph must be greater than 0"
        super().__init__(num_nodes=num_nodes, validate_args=validate_args)
        self.edges_per_node = edges_per_node
        self.num_blocks = num_blocks
        self.damping = damping

    def _one_sample(self):
        """
        Sample a binary adjacency matrix from the underlying distribution.

        Returns:
            A tensor of shape (num_nodes, num_nodes)
        """
        get_edges_per_node = np.random.choice(self.edges_per_node)
        get_num_block = np.random.choice(self.num_blocks)
        get_damping = np.random.choice(self.damping)

        # sample blocks
        splits = np.sort(np.random.choice(self.num_nodes, size=get_num_block - 1, replace=False))
        blocks = np.split(np.random.permutation(self.num_nodes), splits)
        block_sizes = np.array([b.shape[0] for b in blocks])

        # select p s.t. we get requested edges_per_var in expectation
        block_edges_sampled = (np.outer(block_sizes, block_sizes) - np.diag(block_sizes)) / 2
        relative_block_probs = np.eye(get_num_block) + get_damping * (1 - np.eye(get_num_block))
        n_edges = get_edges_per_node * self.num_nodes
        p = min(0.99, n_edges / np.sum(block_edges_sampled * relative_block_probs))

        # sample graph
        mat_intra = np.random.binomial(n=1, p=p, size=(self.num_nodes, self.num_nodes)).astype(int)  # bernoulli
        mat_inter = np.random.binomial(n=1, p=get_damping * p, size=(self.num_nodes, self.num_nodes)).astype(
            int
        )  # bernoulli

        mat = np.zeros((self.num_nodes, self.num_nodes))
        for i, bi in enumerate(blocks):
            for j, bj in enumerate(blocks):
                mat[np.ix_(bi, bj)] = (mat_intra if i == j else mat_inter)[np.ix_(bi, bj)]

        # make directed
        dag = np.triu(mat, k=1)

        # randomly permute
        perm = np.random.permutation(self.num_nodes).tolist()
        dag = dag[perm, :][:, perm]

        if not ig.Graph.Weighted_Adjacency(dag.tolist()).is_dag():
            raise ValueError("Sampled graph is not a DAG")

        dag = torch.tensor(dag, dtype=torch.float32)

        return dag

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        raise NotImplementedError

    def sample(self, sample_shape: torch.Size = torch.Size()):
        """
        Sample binary adjacency matrices from the underlying distribution.

        Args:
            sample_shape: The shape of the samples to be drawn.

        Returns:
            A tensor of shape (sample_shape, num_nodes, num_nodes)
        """
        num_samples = int(torch.prod(torch.tensor(sample_shape)).item())
        return torch.stack([self._one_sample() for _ in range(num_samples)]).reshape(
            sample_shape + (self.num_nodes, self.num_nodes)
        )

    def entropy(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mean(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def mode(self) -> torch.Tensor:
        raise NotImplementedError

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



from dataclasses import dataclass, field

import torch
from tensordict import TensorDictBase

from causica.datasets.causica_dataset_format import CounterfactualWithEffects, InterventionWithEffects


@dataclass
class CausalDataset:
    """Dataclass for storing a causal dataset.

    Args:
        observations: TensorDictBase containing the observations.
        graph: Adjacency matrix of the causal graph.
        noise: Optional, TensorDictBase containing the noise used for generating the observations.
        counterfactuals: Optional, List of CounterfactualWithEffects.
        interventions: Optional, List of InterventionWithEffects.
    """

    observations: TensorDictBase
    graph: torch.Tensor
    node_names: list[str] = field(init=False)  # the nodes that are neither conditioned nor sampled
    noise: TensorDictBase | None = None
    counterfactuals: list[CounterfactualWithEffects] | None = None
    interventions: list[InterventionWithEffects] | None = None

    def __post_init__(self):
        if len(self.observations.batch_size) != 1:
            raise ValueError(f"Batch size must be a scalar, got {self.observations.batch_size}")
        if self.graph.ndim != 2:
            raise ValueError("Expected graph dimension to be 2")

        if self.graph.shape[-1] != self.graph.shape[-2]:
            raise ValueError("The graph must be a square matrix.")

        if self.graph.shape[-1] != len(list(self.observations.keys())):
            raise ValueError(
                f"Graph shape {self.graph.shape} does not match observations: {list(self.observations.keys())}"
            )

        if self.noise is not None and self.noise.batch_size != self.observations.batch_size:
            raise ValueError("Noise batch size does not match observations batch size")

        if self.noise is not None and set(self.noise.keys()) != set(self.observations.keys()):
            raise ValueError("Noise keys do not match observations keys")

        if self.counterfactuals is not None and not all(
            set(cf[0].factual_data.keys()) == set(self.observations.keys()) for cf in self.counterfactuals
        ):
            raise ValueError("Some counterfactual factual data keys do not match observations keys")

        if self.interventions is not None and not all(
            set(intervention[0].intervention_data.keys()) == set(self.observations.keys())
            for intervention in self.interventions
        ):
            raise ValueError("Some intervention data keys do not match observations keys")

        self.node_names = list(self.observations.keys())



from causica.datasets.causica_dataset_format.load import (
    CAUSICA_DATASETS_PATH,
    CounterfactualWithEffects,
    DataEnum,
    InterventionWithEffects,
    Variable,
    VariablesMetadata,
    get_group_idxs,
    get_group_names,
    get_group_variable_names,
    get_name_to_idx,
    load_data,
    tensordict_from_variables_metadata,
    tensordict_to_tensor,
)
from causica.datasets.causica_dataset_format.save import save_data, save_dataset



from typing import Optional

import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import FunctionalRelationships


class LinearFunctionalRelationships(FunctionalRelationships):
    """
    A simple linear functional relationship.
    """

    def __init__(
        self,
        shapes: dict[str, torch.Size],
        initial_linear_coefficient_matrix: torch.Tensor,
        initial_bias: Optional[torch.Tensor] = None,
        trainable: bool = False,
    ) -> None:
        """
        Args:
            shapes: Dict of node shapes (how many dimensions a variable has)
                Order corresponds to the order in graph(s).
            initial_linear_coefficient_matrix: the linear coefficients [output_shape, output_shape]
            initial_bias: Optional, None or a tensor containing the bias [output_shape,]
            trainable: whether the coefficient matrix should be learnable
        """
        super().__init__(shapes=shapes)

        shape = self.tensor_to_td.output_shape
        assert initial_linear_coefficient_matrix.shape == (shape, shape)
        if initial_bias is not None:
            assert initial_bias.shape[0] == shape
        else:
            initial_bias = torch.zeros(shape)

        self.bias = torch.nn.Parameter(initial_bias, requires_grad=trainable)
        self.linear_coefficients = torch.nn.Parameter(initial_linear_coefficient_matrix, requires_grad=trainable)

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """
        Args:
            samples: Batched inputs, size batch_size_x + (concatenated_shape).
            graphs: Weighted adjacency matrix, size batch_size_g + (n, n)
        Returns:
            A Dict of tensors of shape batch_shape_x + batch_shape_g + (concatenated_shape)
        """
        return self.tensor_to_td(self.linear_map(self.tensor_to_td.inv(samples), graphs))

    def linear_map(self, samples: torch.Tensor, graph: torch.Tensor) -> torch.Tensor:
        """
        Applies the linear function to a concatenated tensor of samples.

        Args:
            samples: tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
            graph: tensor of shape batch_shape_g + [n_nodes, n_nodes]
        Returns:
            tensor of shape batch_shape_x + batch_shape_f + batch_shape_g + [n_cols]
        """
        batch_shape_samples = samples.shape[:-1]
        batch_shape_g = graph.shape[:-2]
        batch_shape_x_f = batch_shape_samples[: -len(batch_shape_g)]
        if len(batch_shape_g) > 0 and batch_shape_samples[-len(batch_shape_g) :] != batch_shape_g:
            raise ValueError(
                f"Batch shape of samples and graph must match but got {batch_shape_samples} and {batch_shape_g}"
            )

        masked_graph = torch.einsum("ji,...jk,kl->...il", self.stacked_key_masks, graph, self.stacked_key_masks)

        graph_broad = masked_graph.expand(*(batch_shape_x_f + (-1,) * len(masked_graph.shape)))

        return self.bias + torch.einsum("...i,...ij->...j", samples, graph_broad * self.linear_coefficients)



import math
from typing import Optional

import torch
from torch import nn


def cost_matrix_lp(x_1: torch.Tensor, y_1: torch.Tensor, power: int = 2) -> torch.Tensor:
    """
    Returns the matrix of $|x_1-y_1|^p$.

    Args:
        x_1: expected shape: (*, total_nodes, dim_nodes)
        y_1: expected shape: (*, batch_size, total_nodes, dim_nodes)

    Returns:
        output: expected shape: (*, total_nodes, total_nodes)
    """
    x_col = x_1.unsqueeze(-2)
    y_lin = y_1.unsqueeze(-3)
    return torch.sum((torch.abs(x_col - y_lin)) ** power, -1)


def compute_score_attention(
    queries: torch.Tensor,
    keys: torch.Tensor,
    d_k: float,
    mask: Optional[torch.Tensor] = None,
    cost_type: str = "dot_product",
) -> torch.Tensor:
    """
    Compute the cost matrix between queries and keys.

    Args:
        queries: expected shape: (*, total_nodes, dim_nodes)
        keys: expected shape: (*, total_nodes, dim_nodes)
        d_k (float): rescaling factor
        mask: expected shape: (*, total_nodes, total_nodes)
        cost_type (str): dot_product or l2

    Returns:
        output: expected shape: (*, total_nodes, total_nodes)
    """

    if cost_type == "l2":
        attn_scores = cost_matrix_lp(queries, keys, power=2) / math.sqrt(d_k)
    elif cost_type == "dot_product":
        attn_scores = torch.matmul(queries, keys.transpose(-2, -1)) / math.sqrt(d_k)
    else:
        raise ValueError(f"cost_type {cost_type} not recognized")

    if mask is not None:
        val_masking = -1e4 if attn_scores.dtype == torch.float16 else -1e9
        attn_scores = attn_scores.masked_fill(mask == 0, val_masking)

    return attn_scores


def scaled_dot_product_attention(
    queries: torch.Tensor,
    keys: torch.Tensor,
    d_k: float,
    mask: Optional[torch.Tensor] = None,
    cost_type: str = "dot_product",
) -> torch.Tensor:

    """
    Compute the standard attention matrix

    Args:
        queries: expected shape: (*, total_nodes, dim_nodes)
        keys: expected shape: (*, total_nodes, dim_nodes)
        d_k (float): rescaling factor
        mask: expected shape: (*, total_nodes, total_nodes)
        cost_type (str): dot_product or l2

    Returns:
        output: expected shape: (*, total_nodes, total_nodes)

    """

    attn_scores = compute_score_attention(queries, keys, d_k, mask=mask, cost_type=cost_type)
    return torch.softmax(attn_scores, dim=-1)


def causal_scaled_dot_product_attention(
    queries: torch.Tensor,
    keys: torch.Tensor,
    d_k: float,
    mask: Optional[torch.Tensor] = None,
    cost_type: str = "dot_product",
) -> torch.Tensor:
    r"""
    Partial Optimal Transport problem

    Here we are solving: $\min \langle C, P \rangle$ -\espilon H(P)$ s.t. $P \leq 1$
    where $C$ is the cost matrix, $P$ is the transport matrix, and $H$ is the entropy.

    Args:
        queries: expected shape: (*, total_nodes, dim_nodes)
        keys: expected shape: (*, total_nodes, dim_nodes)
        d_k (float): rescaling factor
        mask: expected shape: (*, total_nodes, total_nodes)
        cost_type (str): dot_product or l2

    Returns:
        output: expected shape: (*, total_nodes, total_nodes)

    """
    attn_scores = compute_score_attention(queries, keys, d_k, mask=mask, cost_type=cost_type)

    max_val = torch.max(attn_scores, dim=-1, keepdim=True)[0]
    attn_scores_rescaled = attn_scores - max_val
    attn_probs_rescaled = torch.exp(attn_scores_rescaled)
    den_rescaled = torch.sum(attn_probs_rescaled, dim=-1, keepdim=True)
    den_rescaled = torch.max(den_rescaled, torch.exp(-max_val))

    return attn_probs_rescaled / den_rescaled


class MultiHeadAttention(nn.Module):
    r"""
    Multi-Head Attention module

    Attention type: attn_type = causal, causal_fixed, diagonal, standard, linear

    In attn_type is "causal_fixed", the attention matrix is fixed
    to be the same for all inputs. In that case, the attention matrix is fixed for any inputs.

    If attn_type is "causal", the attention matrix depends on the input.

    """

    def __init__(
        self,
        d_model: int,
        num_heads: int,
        max_seq_length: int,
        dim_key: Optional[int] = None,
        attn_type: str = "causal",
        cost_type: str = "dot_product",
    ):
        super().__init__()
        self.d_model = d_model
        self.num_heads = num_heads
        self.attn_type = attn_type
        self.cost_type = cost_type

        self.dim_latent = num_heads * dim_key if dim_key is not None else d_model
        assert self.dim_latent % num_heads == 0, "dim_latent must be divisible by num_heads"
        self.d_k = self.dim_latent // num_heads

        self.w_q = nn.Linear(d_model, self.dim_latent)
        self.w_k = nn.Linear(d_model, self.dim_latent)
        self.w_v = nn.Linear(d_model, self.dim_latent)
        self.w_o = nn.Linear(self.dim_latent, d_model)

        if self.attn_type == "standard":
            self.attn = scaled_dot_product_attention
        elif self.attn_type == "linear":
            self.attn = compute_score_attention
        else:
            self.position: torch.Tensor
            self.register_buffer("position", torch.arange(0, max_seq_length, dtype=torch.long))
            self.emb = nn.Embedding(max_seq_length, d_model)
            self.attn = causal_scaled_dot_product_attention

    def split_heads(self, inp: torch.Tensor) -> torch.Tensor:
        """
        Args:
            inp: expected shape (*, max_seq_length, d_model)

        Returns:
            out: expected shape: (*, num_heads, max_seq_length, d_model // num_heads)

        """
        all_dims = inp.size()
        inp = inp.view(*all_dims[:-1], self.num_heads, self.d_k)
        return inp.transpose(-3, -2)

    def combine_heads(self, inp: torch.Tensor) -> torch.Tensor:
        """
        Args:
            inp: expected shape (*, num_heads, max_seq_length, d_model // num_heads)

        Returns:
            out: expected shape: (*, max_seq_length, d_model)

        """
        inp = inp.transpose(-3, -2).contiguous()
        all_dims = inp.size()
        return inp.view(*all_dims[:-2], self.dim_latent)

    def compute_attn(
        self, queries: torch.Tensor, keys: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Args:
            queries: expected shape: (batch_size, max_seq_length, d_model)
            keys: expected shape: (batch_size, max_seq_length, d_model)

        Returns:
            output: expected shape: (batch_size, num_heads, max_seq_length, max_seq_length)

        """

        if self.attn_type == "causal_fixed":
            keys = self.emb(self.position).repeat(keys.size(0), 1, 1)
            queries = self.emb(self.position).repeat(queries.size(0), 1, 1)

        queries = self.split_heads(self.w_q(queries))
        keys = self.split_heads(self.w_k(keys))
        return self.attn(queries, keys, self.d_k, mask=mask, cost_type=self.cost_type)

    def compute_cost(
        self, queries: torch.Tensor, keys: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Args:
            queries: expected shape: (batch_size, max_seq_length, d_model)
            keys: expected shape: (batch_size, max_seq_length, d_model)

        Returns:
            output: expected shape: (batch_size, num_heads, max_seq_length, max_seq_length)

        """
        if self.attn_type == "causal_fixed":
            keys = self.emb(self.position).repeat(keys.size(0), 1, 1)
            queries = self.emb(self.position).repeat(queries.size(0), 1, 1)

        queries = self.split_heads(self.w_q(queries))
        keys = self.split_heads(self.w_k(keys))
        return compute_score_attention(queries, keys, self.d_k, mask=mask, cost_type=self.cost_type)

    def forward(
        self, queries: torch.Tensor, keys: torch.Tensor, values: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Args:
            queries: expected shape: (batch_size, max_seq_length, d_model)
            keys: expected shape: (batch_size, max_seq_length, d_model)
            values: expected shape: (batch_size, max_seq_length, d_model)

        Returns:
            output: expected shape: (batch_size, max_seq_length, max_seq_length)

        """
        values = self.split_heads(self.w_v(values))
        attn_probs = self.compute_attn(queries, keys, mask=mask)
        attn_output = attn_probs @ values

        return self.w_o(self.combine_heads(attn_output))



"""Lightning Classes for loading data in the default format used in Azure Blob Storage."""
import functools
import os
from collections import defaultdict
from functools import partial
from typing import Any, Iterable, Optional, Union

import torch
from tensordict import TensorDict, TensorDictBase
from torch.utils.data import DataLoader

from causica.datasets.causica_dataset_format import CAUSICA_DATASETS_PATH, DataEnum, VariablesMetadata, load_data
from causica.datasets.interventional_data import CounterfactualData, InterventionData
from causica.datasets.normalization import (
    FitNormalizerType,
    Normalizer,
    chain_normalizers,
    fit_log_normalizer,
    fit_standardizer,
)
from causica.datasets.tensordict_utils import identity, tensordict_shapes
from causica.datasets.variable_types import VariableTypeEnum
from causica.distributions.transforms import JointTransformModule
from causica.lightning.data_modules.deci_data_module import DECIDataModule


class VariableSpecDataModule(DECIDataModule):
    """
    Loads training and test data from fully specified paths for `variables.json` formatted data.

    This format assumes the `variables.json` specified all metadata, and that the corresponding CSV files do not have
    any header rows.

    Note:
        Uses `fsspec` to load the data in the paths. To load from a cloud storage location, make sure the relevant
        `fsspec` plugin is available and provide its corresponding scheme. Provide **storage_options for authentication.
        E.g., to load from an Azure storage account, install `adlfs` and use `az://container@storage_account/path`.

    """

    def __init__(
        self,
        root_path: str,
        batch_size: int = 128,
        dataset_name: str = "anonymous_dataset",
        standardize: Union[bool, Iterable[str]] = False,
        log_normalize: Union[bool, Iterable[str]] = False,
        exclude_standardization: Iterable[str] = tuple(),
        exclude_log_normalization: Iterable[str] = tuple(),
        default_offset: float = 1.0,
        log_normalize_min_margin: float = 0.0,
        fit_normalizer_on_test_sets: bool = False,
        load_counterfactual: bool = False,
        load_interventional: bool = False,
        load_validation: bool = False,
        **storage_options: Any,
    ):
        """
        Args:
            root_path: Path to directory with causal data
            batch_size: Batch size for training and test data.
            dataset_name: A name for the dataset
            standardize: Whether to standardize the data or not. It is applied to all continuous variables if True, or applied to those
                variables specifed in normalize except those specified in
                `exclude_normalization`. The standardizer is column-wise: (x_i-mean_i)/std_i for ith column.
                If both standardize and log_normalize are True, log_normalize will be applied first.
            log_normalize: Whether to log normalize the data. If True, it will log normalize all continuous variables.
                Or it will be applied to those variables specified in log_normalize except those specified in `exclude_log_normalization`.
                The operation is
                log(x_i - min_i * (min_i < 0) + min_i + (max_i - min_i) * min_margin + offset). Also see the reference
                in datasets.normalization.LogTransform. If both standardize and log_normalize are True,
                log_normalize will be applied first.
            exclude_standardization: Which variables to exclude from standardization
            exclude_log_normalization: Which variables to exclude from log normalization
            default_offset: Default offset for log normalization.
            log_normalize_min_margin: Minimum margin for log normalization.
            fit_normalizer_on_test_sets: Whether to normalize all data. If True, normalization will be fitted on all data splits.
            load_counterfactual: Whether counterfactual data should be loaded
            load_interventional: Whether interventional data should be loaded
            load_validation: Whether to load the validation dataset
            **storage_options: Storage options forwarded to `fsspec` when loading files.
        """
        super().__init__()
        self.batch_size = batch_size
        self._dataset_name = dataset_name
        self.root_path = root_path
        self.batch_size = batch_size
        self.storage_options = storage_options
        self.standardize = standardize
        self.log_normalize = log_normalize
        self.exclude_standardization = set(exclude_standardization)
        self.exclude_log_normalization = set(exclude_log_normalization)
        self.load_counterfactual = load_counterfactual
        self.load_interventional = load_interventional
        self.load_validation = load_validation
        self.default_offset = default_offset
        self.log_normalize_min_margin = log_normalize_min_margin
        self.fit_normalizer_on_test_sets = fit_normalizer_on_test_sets

        self.use_normalizer = standardize or log_normalize
        self.normalizer: Optional[Normalizer] = None
        self._dataset_train: TensorDictBase
        self._dataset_test: TensorDictBase
        self._dataset_valid: TensorDictBase
        self.true_adj: torch.Tensor
        self.save_hyperparameters()

    @property
    def variable_shapes(self) -> dict[str, torch.Size]:
        return _check_exists(self, "_variable_shapes")

    @property
    def variable_types(self) -> dict[str, VariableTypeEnum]:
        return _check_exists(self, "_variable_types")

    @property
    def column_names(self) -> dict[str, list[str]]:
        return _check_exists(self, "_column_names")

    @property
    def dataset_train(self) -> TensorDict:
        return _check_exists(self, "_dataset_train")

    @property
    def dataset_test(self) -> TensorDict:
        return _check_exists(self, "_dataset_test")

    @property
    def dataset_valid(self) -> TensorDict:
        return _check_exists(self, "_dataset_valid")

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    def create_normalizer(self, normalization_variables: set[str]) -> FitNormalizerType:
        """Return a fitting method for a sequence of normalizers.

        This function is used to return a fitting method for a sequence of normalizers (e.g. log_normalize and standardize).
        The variables for each normalizer is normalization_variables - exclude_normalizer, where normalization_variables is
        a large set of variables (e.g. all continuous variables) and exclude_normalizer is a set of variables that should be excluded specific
        to this normalizer (e.g. exclude_log_normalization and exclude_standardization).

        Args:
            normalization_variables: A larger set of variables to be normalized. It should at least contain the union of variables from all normalizers.

        Returns:
            Return a fitting method for a sequence of normalizers.
        """
        preprocessing: list[FitNormalizerType] = []
        # Setup different separate normalizers
        if self.log_normalize:
            log_normalize_keys = normalization_variables - self.exclude_log_normalization
            if isinstance(self.log_normalize, Iterable):
                log_normalize_keys = set(self.log_normalize) - self.exclude_log_normalization

            fit_log_normalizer_with_key = functools.partial(
                fit_log_normalizer,
                default_offset=self.default_offset,
                min_margin=self.log_normalize_min_margin,
                keys=list(log_normalize_keys),
            )
            preprocessing.append(fit_log_normalizer_with_key)

        if self.standardize:
            standardize_keys = normalization_variables - self.exclude_standardization
            if isinstance(self.standardize, Iterable):
                standardize_keys = set(self.standardize) - self.exclude_standardization

            fit_standardizer_with_key = functools.partial(
                fit_standardizer,
                keys=list(standardize_keys),
            )
            preprocessing.append(fit_standardizer_with_key)

        return chain_normalizers(*preprocessing)

    def _load_all_data(self, variables_metadata: VariablesMetadata):
        _load_data = partial(
            load_data, root_path=self.root_path, variables_metadata=variables_metadata, **self.storage_options
        )

        dataset_train = _load_data(data_enum=DataEnum.TRAIN)
        dataset_test = _load_data(data_enum=DataEnum.TEST)
        if self.load_validation:
            dataset_valid = _load_data(data_enum=DataEnum.VALIDATION)
            assert isinstance(dataset_valid, TensorDict)
            self._dataset_valid = dataset_valid
        true_adj = _load_data(data_enum=DataEnum.TRUE_ADJACENCY)
        assert isinstance(dataset_train, TensorDict)
        assert isinstance(dataset_test, TensorDict)
        assert isinstance(true_adj, torch.Tensor)
        self._dataset_train = dataset_train
        self._dataset_test = dataset_test
        self.true_adj = true_adj

        self.interventions = []
        if self.load_interventional:
            self.interventions = _load_data(data_enum=DataEnum.INTERVENTIONS)

        self.counterfactuals = []
        if self.load_counterfactual:
            self.counterfactuals = _load_data(data_enum=DataEnum.COUNTERFACTUALS)

    def prepare_data(self):
        # WARNING: Do not remove partial here. For some reason, if `load_data` is called directly from inside this data
        # module without being wrapped in partial, it's first optional `variables_metadata` argument is added to the
        # config values of this class. Upon init through the CLI, this argument then becomes interpreted as a value in
        # `storage_options`, whereby other calls to `load_data` will fail due to having two keyword arguments named
        # `variables_metadata`.
        #
        # I.e. if `load_data` is called directly from here, the lightning command line with this class as a data module
        # and `--print_config` produces:
        # ...
        # data:
        #  class_path: causica.lightning.data_modules.VariableSpecDataModule
        #  init_args:
        #    ...
        #    variables_metadata: null
        _load_data = partial(load_data, root_path=self.root_path, **self.storage_options)  # type: ignore
        variables_metadata: VariablesMetadata = _load_data(data_enum=DataEnum.VARIABLES_JSON)  # type: ignore

        self._load_all_data(variables_metadata)

        train_keys = set(self._dataset_train.keys())
        test_keys = set(self._dataset_test.keys())
        assert (
            train_keys == test_keys
        ), f"node_names for the training and test data must match. Diff: {train_keys.symmetric_difference(test_keys)}"
        self._variable_shapes = tensordict_shapes(self._dataset_train)
        self._variable_types = {var.group_name: var.type for var in variables_metadata.variables}
        self._column_names = defaultdict(list)
        for variable in variables_metadata.variables:
            self._column_names[variable.group_name].append(variable.name)

        if self.use_normalizer:
            # Only applied to continuous variables
            normalization_variables = {k for k, v in self._variable_types.items() if v == VariableTypeEnum.CONTINUOUS}
            normalization_sets = [self._dataset_train, self._dataset_test]
            if self.load_validation:
                normalization_sets.append(self._dataset_valid)
            normalization_data = (
                torch.cat(normalization_sets) if self.fit_normalizer_on_test_sets else self._dataset_train
            )
            self.normalizer = self.create_normalizer(normalization_variables)(
                normalization_data.select(*normalization_variables)
            )
            self.normalize_data()
        else:
            self.normalizer = JointTransformModule({})

    def normalize_data(self):
        self._dataset_train = self.normalizer(self._dataset_train)
        self._dataset_test = self.normalizer(self._dataset_test)
        if self._dataset_test.apply(torch.isnan).any():
            raise ValueError(
                "NaN values found in the test data after normalization. Consider changing the normalization parameters."
            )
        if self.load_validation:
            self._dataset_valid = self.normalizer(self._dataset_valid)
            if self._dataset_valid.apply(torch.isnan).any():
                raise ValueError(
                    "NaN values found in the validation data after normalization. Consider changing the normalization parameters."
                )

        if self.load_interventional:
            for intervention in self.interventions:
                for i in intervention:
                    if isinstance(i, InterventionData):
                        i.intervention_data = self.normalizer(i.intervention_data)
                        i.intervention_values = self.normalizer(i.intervention_values)
                        i.condition_values = self.normalizer(i.condition_values)

        if self.load_counterfactual:
            for cf in self.counterfactuals:
                for c in cf:
                    if isinstance(c, CounterfactualData):
                        c.counterfactual_data = self.normalizer(c.counterfactual_data)
                        c.factual_data = self.normalizer(c.factual_data)
                        c.intervention_values = self.normalizer(c.intervention_values)

    def train_dataloader(self):
        return DataLoader(
            dataset=self.dataset_train,
            collate_fn=identity,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
        )

    def test_dataloader(self):
        test_dataloader = DataLoader(dataset=self.dataset_test, collate_fn=identity, batch_size=self.batch_size)

        dataloader_list = [
            test_dataloader,
            DataLoader(dataset=self.true_adj[None, ...]),
            DataLoader(dataset=self.interventions, collate_fn=identity, batch_size=None),
            DataLoader(dataset=self.counterfactuals, collate_fn=identity, batch_size=None),
        ]
        return dataloader_list


class CSuiteDataModule(VariableSpecDataModule):
    """CSuite data module for loading by the datasets name.

    Args:
        dataset_name: Name of the dataset to load.
        batch_size: Batch size for all datasets.
        dataset_path: Path to CSuite dataset mirror.
    """

    def __init__(
        self,
        dataset_name: str,
        batch_size: int = 128,
        dataset_path: str = CAUSICA_DATASETS_PATH,
        load_counterfactual: bool = False,
        load_interventional: bool = False,
        standardize: Union[bool, Iterable[str]] = False,
    ):
        super().__init__(
            root_path=os.path.join(dataset_path, dataset_name),
            batch_size=batch_size,
            dataset_name=dataset_name,
            load_counterfactual=load_counterfactual,
            load_interventional=load_interventional,
            standardize=standardize,
        )


def _check_exists(obj: Any, attribute_name: str):
    """Check if an attribute exists otherwise print a message."""
    try:
        return getattr(obj, attribute_name)
    except AttributeError as exc:
        display_string = attribute_name.replace("_", " ").strip()
        raise ValueError(f"Tried to get {display_string} before data was downloaded.") from exc



from causica.nn.fip.attention_layer import MultiHeadAttention
from causica.nn.fip.avici_encoder import AmortizedEncoder
from causica.nn.fip.decoder_layer import AmortizedDecoderLayer, CausalDecoderLayer
from causica.nn.fip.embeddings import Encoding, PositionWiseFeedForward



import logging
from dataclasses import asdict, is_dataclass
from typing import Any, Mapping, Optional, Sequence, Union

import fsspec
import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.utilities.types import STEP_OUTPUT
from tensordict import TensorDict

from causica.datasets.causica_dataset_format import CounterfactualWithEffects, InterventionWithEffects
from causica.distributions import (
    AdjacencyDistribution,
    ConstrainedAdjacency,
    DistributionModule,
    ENCOAdjacencyDistributionModule,
    ExpertGraphContainer,
    GibbsDAGPrior,
    JointNoiseModule,
    create_noise_modules,
)
from causica.distributions.noise.joint import ContinuousNoiseDist
from causica.functional_relationships import DECIEmbedFunctionalRelationships
from causica.graph.dag_constraint import calculate_dagness
from causica.graph.evaluation_metrics import adjacency_f1, orientation_f1
from causica.lightning.callbacks import AuglagLRCallback
from causica.lightning.data_modules.deci_data_module import DECIDataModule
from causica.lightning.modules.variable_spec_module import VariableSpecModule
from causica.sem.sem_distribution import SEMDistributionModule
from causica.sem.structural_equation_model import SEM
from causica.training.auglag import AugLagLossCalculator, AugLagLR, AugLagLRConfig
from causica.training.evaluation import (
    eval_ate_rmse,
    eval_intervention_likelihoods,
    eval_ite_rmse,
    list_logsumexp,
    list_mean,
)

logging.basicConfig(level=logging.INFO)

NUM_GRAPH_SAMPLES = 100
NUM_ATE_ITE_SEMS = 10


class DECIModule(VariableSpecModule):
    """A PyTorch Lightning Module for running a DECI model."""

    def __init__(
        self,
        noise_dist: ContinuousNoiseDist = ContinuousNoiseDist.SPLINE,
        embedding_size: int = 32,
        out_dim_g: int = 32,
        num_layers_g: int = 2,
        num_layers_zeta: int = 2,
        init_alpha: float = 0.0,
        init_rho: float = 1.0,
        prior_sparsity_lambda: float = 0.05,
        gumbel_temp: float = 0.25,
        auglag_config: Optional[AugLagLRConfig] = None,
        expert_graph_container: Optional[ExpertGraphContainer] = None,
        constraint_matrix_path: Optional[str] = None,
        disable_auglag_epochs: Optional[int] = None,
        test_metric_prefix: Optional[str] = "eval",
    ):
        """DECI Module
        Args:
            noise_dist: The type of the noise distribution to use. The noise type will be used for each variable in the
                SEM.
            embedding_size: The size of the embedding used in the functional relationships. This specifies the
                flexibility of the trainable embeddings for each node.
            out_dim_g: The size of the output of the functional relationships encoder g.
            num_layers_g: The number of layers in the functional relationships encoder g.
            num_layers_zeta: The number of layers in the decoder zeta.
            init_alpha: The initial value of the augmented lagrangian parameter alpha. The augmented lagrangian is
                \rho h(G)^2 + \alpha h(G), where h(G) is the DAG constraint.
            init_rho: The initial value of the augmented lagrangian parameter rho.
            prior_sparsity_lambda: The sparsity coefficient in the graph prior. Larger value prefers sparser graphs.
            gumbel_temp: The temperature used in the gumbel softmax relaxation.
            auglag_config: The configuration for the augmented lagrangian scheduler.
            expert_graph_container: The expert graph container to use. It contains the prior knowledge graph that
                the inferred graph should be close to.
            constraint_matrix_path: The path to the constraint matrix. The constraint matrix specifies the edge
                directions that are prohibited or required.
            disable_auglag_epochs: The number of epochs to disable the augmented lagrangian for.
            test_metric_prefix: Prefix used for logging. The logged metric name will be
                {test_metric_prefix}/{metric_name}.
        """
        super().__init__()
        self.auglag_config = auglag_config if auglag_config is not None else AugLagLRConfig()
        self.disable_auglag_epochs = disable_auglag_epochs
        self.lr_scheduler = AugLagLR(config=self.auglag_config)
        self.constraint_matrix_path = constraint_matrix_path
        self.constraint_matrix: Optional[torch.Tensor] = None

        self.embedding_size = embedding_size
        self.out_dim_g = out_dim_g
        self.num_layers_g = num_layers_g
        self.num_layers_zeta = num_layers_zeta

        self.expert_graph_container: Optional[ExpertGraphContainer] = expert_graph_container

        self.gumbel_temp = gumbel_temp
        self.is_setup = False
        self.noise_dist = noise_dist
        self.prior_sparsity_lambda = prior_sparsity_lambda
        self.test_metric_prefix = test_metric_prefix

        self.auglag_loss: AugLagLossCalculator = AugLagLossCalculator(init_alpha=init_alpha, init_rho=init_rho)

        # Inferred once the datamodule is available using `self.infer_missing_state_from_dataset()`
        self.num_samples = None
        self.variable_group_shapes = None
        self.variable_types = None

        self.save_hyperparameters(logger=False)

    def on_train_start(self) -> None:
        super().on_train_start()

        # Log a version of hparams that are easier to handle for loggers, by converting dataclasses to dicts
        if self.logger is not None:
            hyperparams = {
                k: asdict(v) if is_dataclass(v) and not isinstance(v, type) else v
                for k, v in self.hparams.items()
                if not isinstance(v, torch.nn.Module)
            }
            self.logger.log_hyperparams(hyperparams)

    def prepare_data(self) -> None:
        """Set the constraint matrix (if necessary)."""
        if self.constraint_matrix_path:
            with fsspec.open(self.constraint_matrix_path) as f:
                if self.constraint_matrix_path.endswith("npy"):
                    constraint_matrix = np.load(f)
                else:
                    raise ValueError(f"Unsupported constraint matrix format: {self.constraint_matrix_path}")
            self.constraint_matrix = torch.tensor(constraint_matrix)
        return super().prepare_data()

    def infer_missing_state_from_dataset(self):
        dataset_defined_members = [
            self.num_samples,
            self.variable_group_shapes,
            self.variable_types,
        ]
        if any(member is None for member in dataset_defined_members):
            datamodule = getattr(self.trainer, "datamodule", None)
            if not isinstance(datamodule, DECIDataModule):
                raise TypeError(
                    f"Incompatible data module {datamodule}, requires a DECIDataModule but is "
                    f"{type(datamodule).mro()}"
                )
            if self.num_samples is None:
                self.num_samples = len(datamodule.dataset_train)
            if self.variable_group_shapes is None:
                self.variable_group_shapes = datamodule.variable_shapes
            if self.variable_types is None:
                self.variable_types = datamodule.variable_types

    def setup(self, stage: Optional[str] = None):

        _ = stage
        if self.is_setup:
            return  # Already setup

        self.infer_missing_state_from_dataset()
        assert self.num_samples is not None
        assert self.variable_group_shapes is not None
        assert self.variable_types is not None

        num_nodes = len(self.variable_group_shapes)

        adjacency_dist: DistributionModule[AdjacencyDistribution] = ENCOAdjacencyDistributionModule(num_nodes)
        if self.constraint_matrix is not None:
            adjacency_dist = ConstrainedAdjacency(adjacency_dist, self.constraint_matrix)

        functional_relationships = DECIEmbedFunctionalRelationships(
            shapes=self.variable_group_shapes,
            embedding_size=self.embedding_size,
            out_dim_g=self.out_dim_g,
            num_layers_g=self.num_layers_g,
            num_layers_zeta=self.num_layers_zeta,
        )
        noise_submodules = create_noise_modules(self.variable_group_shapes, self.variable_types, self.noise_dist)
        noise_module = JointNoiseModule(noise_submodules)
        self.sem_module: SEMDistributionModule = SEMDistributionModule(
            adjacency_dist, functional_relationships, noise_module
        )
        self.prior: GibbsDAGPrior = GibbsDAGPrior(
            num_nodes=num_nodes,
            sparsity_lambda=self.prior_sparsity_lambda,
            expert_graph_container=self.expert_graph_container,
        )
        self.is_setup = True

    def training_step(self, *args, **kwargs) -> STEP_OUTPUT:
        _ = kwargs
        batch, *_ = args
        batch = batch.apply(lambda t: t.to(torch.float32, non_blocking=True))

        sem_distribution = self.sem_module()
        sem, *_ = sem_distribution.relaxed_sample(torch.Size([]), temperature=self.gumbel_temp)  # soft sample

        batch_log_prob = sem.log_prob(batch).mean()
        sem_distribution_entropy = sem_distribution.entropy()
        prior_term = self.prior.log_prob(sem.graph)
        objective = (-sem_distribution_entropy - prior_term) / self.num_samples - batch_log_prob
        constraint = calculate_dagness(sem.graph)
        step_output = {
            "alpha": self.auglag_loss.alpha,
            "rho": self.auglag_loss.rho,
            "loss": self.auglag_loss(objective, constraint / self.num_samples),
            "batch_log_prob": batch_log_prob,
            "constraint": constraint,
            "num_edges": (sem.graph > 0.0).count_nonzero(),
            "vardist_entropy": sem_distribution_entropy,
            "prior_term": prior_term,
        }
        self.log_dict(step_output, on_epoch=True)  # Only log this on epoch end
        return step_output

    def configure_optimizers(self):
        """Set the learning rates for different sets of parameters."""
        modules = {
            "functional_relationships": self.sem_module.functional_relationships,
            "vardist": self.sem_module.adjacency_module,
            "noise_dist": self.sem_module.noise_module,
        }
        parameter_list = [
            {
                "params": module.parameters(),
                "lr": self.auglag_config.lr_init_dict[name],
                "name": name,
            }
            for name, module in modules.items()
        ]

        # Check that all modules are added to the parameter list
        check_modules = set(modules.values())
        for module in self.parameters(recurse=False):
            assert module in check_modules, f"Module {module} not in module list"

        return torch.optim.Adam(parameter_list)

    def configure_callbacks(self) -> Union[Sequence[pl.Callback], pl.Callback]:
        """Create a callback for the auglag callback."""
        disabled_epochs = set(range(self.disable_auglag_epochs)) if self.disable_auglag_epochs else None
        return [AuglagLRCallback(self.lr_scheduler, log_auglag=True, disabled_epochs=disabled_epochs)]

    def test_step_observational(self, batch: TensorDict, *args, **kwargs):
        """Evaluate the log prob of the model on the test set using multiple graph samples."""
        _, _ = args, kwargs
        batch = batch.apply(lambda t: t.to(torch.float32, non_blocking=True))
        sems = self.sem_module().sample(torch.Size([NUM_GRAPH_SAMPLES]))
        dataset_size = self.trainer.datamodule.dataset_test.batch_size  # type: ignore
        assert len(dataset_size) == 1, "Only one batch size is supported"

        # Estimate log prob for each sample using graph samples and report the mean over graphs
        log_prob_test = list_logsumexp([sem.log_prob(batch) for sem in sems]) - np.log(NUM_GRAPH_SAMPLES)
        self.log(
            f"{self.test_metric_prefix}/test_LL",
            torch.sum(log_prob_test, dim=-1).item() / dataset_size[0],
            reduce_fx=sum,
            add_dataloader_idx=False,
        )

    def test_step_graph(self, true_adj_matrix: torch.Tensor, *args, **kwargs):
        """Evaluate the graph reconstruction performance of the model against the true graph."""
        _, _ = args, kwargs
        sem_samples = self.sem_module().sample(torch.Size([NUM_GRAPH_SAMPLES]))
        graph_samples = [sem.graph for sem in sem_samples]

        adj_f1 = list_mean([adjacency_f1(true_adj_matrix, graph) for graph in graph_samples]).item()
        orient_f1 = list_mean([orientation_f1(true_adj_matrix, graph) for graph in graph_samples]).item()
        self.log(f"{self.test_metric_prefix}/adjacency.f1", adj_f1, add_dataloader_idx=False)
        self.log(f"{self.test_metric_prefix}/orientation.f1", orient_f1, add_dataloader_idx=False)

    def test_step_interventions(self, interventions: InterventionWithEffects, *args, **kwargs):
        """Evaluate the ATE and Interventional log prob performance of the model"""
        _, _ = args, kwargs
        sems_list: list[SEM] = list(self.sem_module().sample(torch.Size([NUM_ATE_ITE_SEMS])))
        interventional_log_prob = eval_intervention_likelihoods(sems_list, interventions)
        self.log(
            f"{self.test_metric_prefix}/Interventional_LL",
            torch.mean(interventional_log_prob).item(),
            add_dataloader_idx=False,
        )

        mean_ate_rmse = eval_ate_rmse(sems_list, interventions)
        self.log(
            f"{self.test_metric_prefix}/ATE_RMSE",
            list_mean(list(mean_ate_rmse.values())).item(),
            add_dataloader_idx=False,
        )

    def test_step_counterfactuals(self, counterfactuals: CounterfactualWithEffects, *args, **kwargs):
        """Evaluate the ITE performance of the model"""
        _, _ = args, kwargs
        sems_list = list(self.sem_module().sample(torch.Size([NUM_ATE_ITE_SEMS])))
        ite_rmse = eval_ite_rmse(sems_list, counterfactuals)
        self.log(
            f"{self.test_metric_prefix}/ITE_RMSE",
            list_mean(list(ite_rmse.values())).item(),
            add_dataloader_idx=False,
        )

    def load_state_dict(self, state_dict: Mapping[str, Any], strict: bool = True):
        # initialise all the parameters, we can
        super().load_state_dict(state_dict, strict=False)
        # setup the model
        self.setup()
        # load the state dict again to fill in the parameters
        return super().load_state_dict(state_dict, strict=strict)

    def set_extra_state(self, state: Any):
        self.variable_group_shapes = state["shapes"]
        self.variable_types = state["types"]
        self.constraint_matrix = state["constraint_matrix"]
        self.num_samples = state["num_samples"]

    def get_extra_state(self) -> Any:
        return {
            "shapes": self.variable_group_shapes,
            "types": self.variable_types,
            "constraint_matrix": self.constraint_matrix,
            "num_samples": self.num_samples,
        }



import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import FunctionalRelationships
from causica.nn import TemporalEmbedNN


class TemporalEmbedFunctionalRelationships(FunctionalRelationships):
    """This is a `FunctionalRelationsips` that wraps the `RhinoEmbedNN` module."""

    def __init__(
        self,
        shapes: dict[str, torch.Size],
        embedding_size: int,
        out_dim_g: int,
        num_layers_g: int,
        num_layers_zeta: int,
        context_length: int,
    ) -> None:
        """
        Args:
            shapes: Dictionary of shapes of the input tensors.
            embedding_size: See `TemporalEmbedNN`.
            out_dim_g: See `TemporalEmbedNN`.
            num_layers_g: See `TemporalEmbedNN`.
            num_layers_zeta: See `TemporalEmbedNN`.
            context_length: See `TemporalEmbedNN`.
        """
        super().__init__(shapes=shapes)
        self.nn = TemporalEmbedNN(
            group_mask=self.stacked_key_masks,
            embedding_size=embedding_size,
            out_dim_l=out_dim_g,
            num_layers_l=num_layers_g,
            num_layers_zeta=num_layers_zeta,
            context_length=context_length,
        )

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        return self.tensor_to_td(self.nn(self.tensor_to_td.inv(samples), graphs))



import abc
from typing import Optional

import torch
import torch.distributions as td
import torch.nn.functional as F
from torch import nn
from torch.distributions.utils import logits_to_probs

from causica.distributions.adjacency.adjacency_distributions import AdjacencyDistribution
from causica.distributions.distribution_module import DistributionModule
from causica.distributions.gumbel_binary import gumbel_softmax_binary


class LaggedAdjacencyDistribution(AdjacencyDistribution, abc.ABC):
    """
    Probability distributions of the lagged adjacency matrix for causal timeseries model.

    The main functionality is similar to `AdjacencyDistribution`. The main differences are that we additional have
    `lag` parameters to indicate the window size of the temporal adjacency matrix. E.g. lag=3
    indicates we only consider previous 3 step of observations. `lags` equals to `context_length` - 1.

    Another difference is that we do not disable the diagonals of the lagged adjacency matries and no dagness constraints as well.

    One single sample from this lagged distribution should have shape (lags, num_nodes, num_nodes). lags must be greater than 0.
    For multple samples, the shape will be sample_shape + batch_shape + (lags, num_nodes, num_nodes)
    """

    def __init__(self, num_nodes: int, lags: int, validate_args: Optional[bool] = None):
        assert lags > 0, "Number of lags must be greater than 0"
        self.lags = lags
        assert num_nodes > 0, "Number of nodes in the graph must be greater than 0"
        self.num_nodes = num_nodes
        event_shape = torch.Size((lags, num_nodes, num_nodes))

        super(AdjacencyDistribution, self).__init__(event_shape=event_shape, validate_args=validate_args)


class TemporalAdjacencyDistribution(td.Distribution):
    """Probability distributions of the temporal adjacency matrix for causal timeseries model.

    Combines an instantaneous AdjacencyDistribution and temporal LaggedAdjacencyDistribution to form a proper temporal adjacency matrix.

    One single sample from this distribution should have shape (context_length, num_nodes, num_nodes)
    For multple samples, the shape will be sample_shape + batch_shape + (context_length, num_nodes, num_nodes)

    For single sample, graph[-1] will be the instantaneous graph and graph[0] will be the last lagged graph (if context_length > 1).
    """

    support = td.constraints.independent(td.constraints.boolean, 1)
    arg_constraints = {}

    def __init__(
        self,
        instantaneous_distribution: AdjacencyDistribution,
        lagged_distribution: Optional[LaggedAdjacencyDistribution],
        validate_args: Optional[bool] = None,
    ):
        """
        Args:
            instantaneous_distribution: The distribution of the instantaneous adjacency matrix
            lagged_distribution: The distribution of the lagged adjacency matrix
            validate_args: Whether to validate the input arguments
        """
        if lagged_distribution is not None:
            self.lags = lagged_distribution.lags
            assert (
                instantaneous_distribution.num_nodes == lagged_distribution.num_nodes
            ), "Number of nodes in the graph must be the same"
        else:
            self.lags = 0
        self.context_length = self.lags + 1
        self.num_nodes = instantaneous_distribution.num_nodes
        event_shape = torch.Size((self.context_length, self.num_nodes, self.num_nodes))
        self.inst_dist = instantaneous_distribution
        self.lagged_dist = lagged_distribution

        super().__init__(event_shape=event_shape, validate_args=validate_args)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """Sample a binary temporal adjacency matrix from the relaxed distribution.

        This relies on the relaxed sampling method of both the instantaneous and lagged distributions, so that gradients
        can flow through. To obtain hard samples that stops gradients, use the `sample` method.

        Args:
            sample_shape: the shape of the samples to return
            temperature: The temperature of the relaxed distribution

        Returns:
            A tensor of shape sample_shape + batch_shape + (context_length, num_nodes, num_nodes)
        """
        # sample inst adj matrix with shape sample_shape + batch_shape + (num_nodes, num_nodes)
        inst_adj_matrix = self.inst_dist.relaxed_sample(sample_shape, temperature)
        if self.lagged_dist is None:
            return inst_adj_matrix.unsqueeze(-3)  # sample_shape + batch_shape + (1, num_nodes, num_nodes)
        # sample lagged adj matrix with shape sample_shape + batch_shape + (lags, num_nodes, num_nodes)
        lagged_adj_matrix = self.lagged_dist.relaxed_sample(sample_shape, temperature)

        return torch.cat(
            [lagged_adj_matrix, inst_adj_matrix.unsqueeze(-3)], dim=-3
        )  # sample_shape + batch_shape + (context_length, num_nodes, num_nodes)

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """Sample a binary temporal adjacency matrix from the underlying distribution.

        The gradients will not flow through this method, use `relaxed_sample` instead.

        Args:
            sample_shape: the shape of the samples to return

        Returns:
            A tensor of shape sample_shape + batch_shape + (context_length, num_nodes, num_nodes)
        """
        inst_adj_matrix = self.inst_dist.sample(sample_shape)  # sample_shape + batch_shape + (num_nodes, num_nodes)
        if self.lagged_dist is None:
            return inst_adj_matrix.unsqueeze(-3)  # sample_shape + batch_shape + (1, num_nodes, num_nodes)

        lagged_adj_matrix = self.lagged_dist.sample(
            sample_shape
        )  # sample_shape + batch_shape + (lags, num_nodes, num_nodes)
        return torch.cat(
            [lagged_adj_matrix, inst_adj_matrix.unsqueeze(-3)], dim=-3
        )  # sample_shape + batch_shape + (context_length, num_nodes, num_nodes)

    def entropy(self) -> torch.Tensor:
        """
        Return the entropy of the underlying temporal adjacency distribution.

        Returns:
            A tensor of shape batch_shape, with the entropy of the distribution
        """
        if self.lagged_dist is None:
            return self.inst_dist.entropy()
        return self.inst_dist.entropy() + self.lagged_dist.entropy()

    @property
    def mean(self) -> torch.Tensor:
        """
        Return the mean of the underlying temporal adjaccency distribution.

        This will be a matrix with all entries in the interval [0, 1].

        Returns:
            A tensor of shape batch_shape + (context_length, num_nodes, num_nodes)
        """
        if self.lagged_dist is None:
            return self.inst_dist.mean.unsqueeze(-3)  # batch_shape + (1, num_nodes, num_nodes)

        return torch.cat(
            [self.lagged_dist.mean, self.inst_dist.mean.unsqueeze(-3)], dim=-3
        )  # batch_shape + (context_length, num_nodes, num_nodes)

    @property
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying temporal adjacency distribution.

        This will be an temporal adjacency matrix.

        Returns:
            A tensor of shape batch_shape + (context_length, num_nodes, num_nodes)
        """
        if self.lagged_dist is None:
            return self.inst_dist.mode.unsqueeze(-3)  # batch_shape + (1, num_nodes, num_nodes)

        return torch.cat(
            [self.lagged_dist.mode, self.inst_dist.mode.unsqueeze(-3)], dim=-3
        )  # batch_shape + (context_length, num_nodes, num_nodes)

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Get the log probability of each tensor from the sample space

        Args:
            value: a binary matrix of shape value_shape + batch_shape + (context_length, n, n)
        Returns:
            A tensor of shape value_shape + batch_shape, with the log probabilities of each tensor in the batch.
        """
        inst_adj_matrix = value[..., -1, :, :]  # value_shape + batch_shape + (num_nodes, num_nodes)
        if self.lagged_dist is None:
            return self.inst_dist.log_prob(inst_adj_matrix)  # value_shape + batch_shape

        lagged_adj_matrix = value[..., :-1, :, :]  # value_shape + batch_shape + (lags, num_nodes, num_nodes)
        return self.inst_dist.log_prob(inst_adj_matrix) + self.lagged_dist.log_prob(lagged_adj_matrix)


class RhinoLaggedAdjacencyDistribution(LaggedAdjacencyDistribution):
    """
    This implements the adjacency distribution for lagged adj matrix.
    """

    arg_constraints = {"logits_edge": td.constraints.real}

    def __init__(
        self,
        logits_edge: torch.Tensor,
        lags: int,
        validate_args: Optional[bool] = None,
    ):
        """
        Args:
            logits_edge: The logits for the edge existence. The shape is a batch_shape + (lags, n, n) tensor.
            lags: The number of lags in the temporal adjacency matrix. It must be positive.
            validate_args: Whether to validate the arguments. Passed to the superclass
        """
        assert lags == logits_edge.shape[-3], "lags must match the number of lags in logits_edge"
        assert lags > 0, "lags must be a positive integer"
        num_nodes = logits_edge.shape[-1]

        if validate_args:
            assert len(logits_edge.shape) >= 3, "logits_exist must be a 3D tensor with shape (lags, n, n)"
            assert logits_edge.shape[-3:] == (lags, num_nodes, num_nodes), "Invalid logits_edge shape"

        self.logits_edge = logits_edge
        self.lags = lags

        super().__init__(num_nodes, lags, validate_args=validate_args)

    def _get_independent_bernoulli_logits(self) -> torch.Tensor:
        """
        Construct the lagged matrix logit(pᵢⱼ).
        The output will have shape [..., lags, node, node]

        """
        # Lagged adjacency logits
        return self.logits_edge  # [..., lags, node, node]

    @staticmethod
    def base_dist(logits: torch.Tensor) -> td.Distribution:
        """A matrix of independent Bernoulli distributions."""
        return td.Independent(td.Bernoulli(logits=logits), 3)

    def relaxed_sample(self, sample_shape: torch.Size = torch.Size(), temperature: float = 0.0) -> torch.Tensor:
        """
        Sample from a relaxed distribution. We use a Gumbel Softmax.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (lags, num_nodes, num_nodes)
        """

        logits = self._get_independent_bernoulli_logits()  # [batch_shape, lags, num_nodes, num_nodes]
        expanded_logits = logits.expand(*(sample_shape + logits.shape))
        samples = gumbel_softmax_binary(
            logits=expanded_logits, tau=temperature, hard=True
        )  # [sample_shape, batch_shape, lags, num_nodes, num_nodes]

        return samples

    def sample(self, sample_shape: torch.Size = torch.Size()) -> torch.Tensor:
        """
        Sample from the underyling independent Bernoulli distribution.

        Gradients will not flow through this method, use relaxed_sample instead.

        Args:
            sample_shape: the shape of the samples to return
        Returns:
            A tensor of shape sample_shape + batch_shape + (lags, num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()  # [batch_shape, lags, num_nodes, num_nodes]
        samples = self.base_dist(logits).sample(
            sample_shape=sample_shape
        )  # [sample_shape, batch_shape, lags, num_nodes, num_nodes]
        return samples

    def entropy(self) -> torch.Tensor:
        """
        Get the entropy of the independent Bernoulli distribution.

        Returns:
            A tensor of shape batch_shape, with the entropy of the distribution
        """
        logits = self._get_independent_bernoulli_logits()
        # shape of entropy is [batch_shape, lags, num_nodes, num_nodes]
        entropy = F.binary_cross_entropy_with_logits(logits, logits_to_probs(logits, is_binary=True), reduction="none")
        # Entropy should not consider diagonal elements for instantaneous adj matrix
        return torch.sum(entropy, dim=(-3, -2, -1))  # batch_shape

    @property
    def mean(self) -> torch.Tensor:
        """
        Return the mean of the underlying independent Bernoulli distribution.

        This will be a matrix with all entries in the interval [0, 1].

        Returns:
            A tensor of shape batch_shape + (lags, num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()
        # shape of mean is [batch_shape, lags, num_nodes, num_nodes]
        return self.base_dist(logits).mean

    @property
    def mode(self) -> torch.Tensor:
        """
        Return the mode of the underlying independent Bernoulli distribution.

        This will be an adjacency matrix.

        Returns:
            A tensor of shape batch_shape + (lags, num_nodes, num_nodes)
        """
        logits = self._get_independent_bernoulli_logits()
        # bernoulli mode can be nan for very small logits, favour sparseness and set to 0
        # shape of mode is [batch_shape, lags, num_nodes, num_nodes]
        mode = torch.nan_to_num(self.base_dist(logits).mode, nan=0.0)

        return mode

    def log_prob(self, value: torch.Tensor) -> torch.Tensor:
        """
        Get the log probability of each tensor from the sample space

        Args:
            value: a binary matrix of shape value_shape + batch_shape + (lags, n, n)
        Returns:
            A tensor of shape value_shape + batch_shape, with the log probabilities of each tensor in the batch.
        """

        logits = self._get_independent_bernoulli_logits()
        full_log_prob = self.base_dist(logits).log_prob(value)  # value_shape + batch_shape

        return full_log_prob


class RhinoLaggedAdjacencyDistributionModule(DistributionModule[RhinoLaggedAdjacencyDistribution]):
    """Represents an `RhinoLaggedAdjacencyDistributionModule` distribution with learnable parameters."""

    def __init__(self, num_nodes: int, lags: int) -> None:
        """
        Args:
            num_nodes: The number of nodes in the graph. It must be positive.
            lags: The number of lags in the temporal adjacency matrix. It must be positive.
        """
        super().__init__()
        self.logits_edge = nn.Parameter(torch.zeros(lags, num_nodes, num_nodes), requires_grad=True)
        self.lags = lags

    def forward(self) -> RhinoLaggedAdjacencyDistribution:
        return RhinoLaggedAdjacencyDistribution(logits_edge=self.logits_edge, lags=self.lags)


class TemporalAdjacencyDistributionModule(DistributionModule[TemporalAdjacencyDistribution]):
    """Represents an `TemporalAdjacencyDistributionModule` distribution with learnable parameters."""

    def __init__(
        self,
        inst_dist_module: DistributionModule[AdjacencyDistribution],
        lagged_dist_module: Optional[DistributionModule[LaggedAdjacencyDistribution]],
    ) -> None:
        """
        Args:
            inst_dist_module: The distribution module of the instantaneous adjacency matrix.
            lagged_dist_module: The distribution module of the lagged adjacency matrix. If None, we assume only
                instantaneous adjacency exists.
        """
        super().__init__()
        self.inst_dist_module = inst_dist_module
        self.lagged_dist_module = lagged_dist_module

    def forward(self) -> TemporalAdjacencyDistribution:
        return TemporalAdjacencyDistribution(
            instantaneous_distribution=self.inst_dist_module(),
            lagged_distribution=self.lagged_dist_module() if self.lagged_dist_module is not None else None,
        )



from causica.distributions.transforms.base import SequentialTransformModule, TransformModule, TypedTransform
from causica.distributions.transforms.joint import JointTransform, JointTransformModule
from causica.distributions.transforms.tensor_to_tensordict import TensorToTensorDictTransform, shapes_to_slices
